package com.withpersona.sdk2.inquiry

import com.withpersona.sdk2.inquiry.internal.InquiryField
import com.withpersona.sdk2.inquiry.internal.InquiryField.BooleanField
import com.withpersona.sdk2.inquiry.internal.InquiryField.ChoicesField
import com.withpersona.sdk2.inquiry.internal.InquiryField.DateField
import com.withpersona.sdk2.inquiry.internal.InquiryField.FloatField
import com.withpersona.sdk2.inquiry.internal.InquiryField.IntegerField
import com.withpersona.sdk2.inquiry.internal.InquiryField.MultiChoicesField
import com.withpersona.sdk2.inquiry.internal.InquiryField.StringField
import java.util.Date

/**
 * Data to prepopulate on the [Inquiry] without requiring individual input
 */
class Fields private constructor(internal val fields: Map<String, InquiryField>) {
  /**
   * Builder for [Fields]
   */
  class Builder {
    private val fields: MutableMap<String, InquiryField> = mutableMapOf()

    /**
     * Set an arbitrary [String] value on the [Inquiry]
     *
     * @param name field name on the [Inquiry] in camelCase (e.g. "emailAddress")
     * @param value [String] value of the field
     * @return [Builder] to continue the builder pattern
     */
    fun field(name: String, value: String): Builder =
      this.apply { fields[name] = StringField(value) }

    /**
     * Set an arbitrary [Int] value on the [Inquiry]
     *
     * @param name field name on the [Inquiry] in camelCase (e.g. "favoriteNumber")
     * @param value [Int] value of the field
     * @return [Builder] to continue the builder pattern
     */
    fun field(name: String, value: Int): Builder = this.apply { fields[name] = IntegerField(value) }

    /**
     * Set an arbitrary [Boolean] value on the [Inquiry]
     *
     * @param name field name on the [Inquiry] in camelCase (e.g. "isLeftHanded")
     * @param value [Boolean] value of the field
     * @return [Builder] to continue the builder pattern
     */
    fun field(name: String, value: Boolean): Builder =
      this.apply { fields[name] = BooleanField(value) }

    /**
     * Set an arbitrary [Float] value on the [Inquiry]
     *
     * @param name field name on the [Inquiry] in camelCase (e.g. "amountAdded")
     * @param value [Float] value of the field
     * @return [Builder] to continue the builder pattern
     */
    fun field(name: String, value: Float): Builder =
      this.apply { fields[name] = FloatField(value) }

    /**
     * Set an arbitrary [Date] value on the [Inquiry]
     *
     * @param name field name on the [Inquiry] in camelCase (e.g. "calendarDate")
     * @param value [Date] value of the field
     * @return [Builder] to continue the builder pattern
     */
    fun field(name: String, value: Date): Builder =
      this.apply { fields[name] = DateField(value.toString()) }

    /**
     * Set an arbitrary [String] value as a Choices field type on the [Inquiry]
     *
     * @param name field name on the [Inquiry] in camelCase (e.g. "options")
     * @param value [String] value of the field
     * @return [Builder] to continue the builder pattern
     */
    fun fieldChoices(name: String, value: String): Builder =
      this.apply { fields[name] = ChoicesField(value) }

    /**
     * Set an arbitrary [Array<String>] value as a MultiChoices field type on the [Inquiry]
     *
     * @param name field name on the [Inquiry] in camelCase (e.g. "options")
     * @param value [Array<String>] value of the field
     * @return [Builder] to continue the builder pattern
     */
    fun fieldMultiChoices(name: String, value: Array<String>): Builder =
      this.apply { fields[name] = MultiChoicesField(value) }

    /**
     * Build the [Fields] object
     *
     * @return assembled [Fields] instance
     */
    fun build(): Fields = Fields(fields)
  }
}