package com.withpersona.sdk2.inquiry

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.os.Bundle
import androidx.activity.result.contract.ActivityResultContract
import androidx.annotation.MainThread
import androidx.core.os.BundleCompat
import com.withpersona.sdk2.inquiry.InquiryField.BooleanField
import com.withpersona.sdk2.inquiry.InquiryField.ChoicesField
import com.withpersona.sdk2.inquiry.InquiryField.DateField
import com.withpersona.sdk2.inquiry.InquiryField.DatetimeField
import com.withpersona.sdk2.inquiry.InquiryField.FloatField
import com.withpersona.sdk2.inquiry.InquiryField.IntegerField
import com.withpersona.sdk2.inquiry.InquiryField.MultiChoicesField
import com.withpersona.sdk2.inquiry.InquiryField.StringField
import com.withpersona.sdk2.inquiry.InquiryField.UnknownField
import com.withpersona.sdk2.inquiry.error_reporting.ExceptionLogger
import com.withpersona.sdk2.inquiry.internal.InquiryField as InternalInquiryField
import com.withpersona.sdk2.inquiry.internal.InquiryFieldsMap
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.ACCOUNT_ID
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.COLLECTED_DATA
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.CONSUME_EXCEPTIONS
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.ENABLE_ERROR_LOGGING
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.ENVIRONMENT
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.ENVIRONMENT_ID
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.ERROR_CODE
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.ERROR_DEBUG_MESSAGE
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.FIELDS_MAP
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.INQUIRY_ID
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.INQUIRY_STATUS
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.LOCALE
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.PERSONA_ACTIVITY_RESULT
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.REFERENCE_ID
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.RETURN_COLLECTED_DATA
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.ROUTING_COUNTRY
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.SESSION_TOKEN
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.SHOULD_AUTO_FALLBACK
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.STATIC_INQUIRY_TEMPLATE
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.Status
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.TEMPLATE_ID
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.TEMPLATE_VERSION
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.THEME
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.THEME_SET_ID
import com.withpersona.sdk2.inquiry.internal.InquiryIntentKeys.USE_SERVER_STYLES
import com.withpersona.sdk2.inquiry.internal.Prefetching
import com.withpersona.sdk2.inquiry.types.INQUIRY_ACTIVITY_NAME
import com.withpersona.sdk2.inquiry.types.InquiryErrorMessages
import com.withpersona.sdk2.inquiry.types.collected_data.CollectedData
import com.withpersona.sdk2.inquiry.types.collected_data.ErrorCode
import java.text.SimpleDateFormat
import java.util.Locale

/**
 * Start a Persona Inquiry flow and parse the response from your Activity.
 *
 * ## Register for result
 * ```
 * val getInquiryResult =
 *   registerForActivityResult(Inquiry.Contract()) { result ->
 *     when (result) {
 *       is InquiryResponse.Complete -> {
 *         // ... completed flow
 *       }
 *       is InquiryResponse.Cancel -> {
 *         // ... abandoned flow
 *       }
 *       is InquiryResponse.Error -> {
 *         // ... something went wrong
 *       }
 *     }
 *   }
 * ```
 *
 * ## Start the flow.
 * ```
 * // Get the template ID from the Dashboard
 * // https://withpersona.com/dashboard/integration
 * val TEMPLATE_ID = "YOUR_TEMPLATE_ID"
 * // ...
 *
 * val inquiry = Inquiry.fromTemplate(TEMPLATE_ID)
 *   .environment(Environment.SANDBOX)
 *   .build()
 * // ...
 *
 * getInquiryResult.launch(inquiry)
 * ```
 */
class Inquiry internal constructor(
  private val templateId: String?,
  private val templateVersion: String?,
  private val inquiryId: String?,
  private val sessionToken: String?,
  private val referenceId: String?,
  private val accountId: String?,
  private val fields: Fields?,
  private val theme: Int?,
  private val environment: Environment?,
  private val environmentId: String?,
  private val enableErrorLogging: Boolean?,
  private val returnCollectedData: Boolean?,
  private val shouldAutoFallback: Boolean?,
  private val useServerStyles: Boolean?,
  private val staticInquiryTemplate: StaticInquiryTemplate?,
  private val themeSetId: String?,
  private val routingCountry: String?,
  private val locale: String?,
  private val consumeExceptions: Boolean?,
) {

  /**
   * Launch the Persona Inquiry.
   *
   * @param activity current Activity that will get the response
   * @param requestCode arbitrary number that will be returned to you in the response
   */
  @MainThread
  @Deprecated("Use registerForActivityResult with the Inquiry.Contract instead.")
  fun start(activity: Activity, requestCode: Int) {
    activity.startActivityForResult(
      toInquiryActivityIntent(activity),
      requestCode,
    )
  }

  @ExperimentalInlineApi
  fun buildInlineInquiry(): InlineInquiryBuilder =
    InlineInquiryBuilder(this)

  private fun toInquiryActivityIntent(context: Context): Intent =
    Intent().apply {
      setClassName(
        context,
        INQUIRY_ACTIVITY_NAME,
      )
      replaceExtras(
        Bundle().apply { addArgumentsToBundle(this) },
      )
    }

  internal fun addArgumentsToBundle(bundle: Bundle) {
    with(bundle) {
      templateId?.let { putString(TEMPLATE_ID, it) }
      templateVersion?.let { putString(TEMPLATE_VERSION, it) }
      inquiryId?.let { putString(INQUIRY_ID, it) }
      sessionToken?.let { putString(SESSION_TOKEN, it) }
      referenceId?.let { putString(REFERENCE_ID, it) }
      accountId?.let { putString(ACCOUNT_ID, it) }
      fields?.let { putParcelable(FIELDS_MAP, InquiryFieldsMap(it.fields)) }
      theme?.let { putInt(THEME, it) }
      staticInquiryTemplate?.let { putParcelable(STATIC_INQUIRY_TEMPLATE, it) }
      environment?.let { putString(ENVIRONMENT, it.name) }
      environmentId?.let { putString(ENVIRONMENT_ID, it) }
      enableErrorLogging?.let { putBoolean(ENABLE_ERROR_LOGGING, it) }
      returnCollectedData?.let { putBoolean(RETURN_COLLECTED_DATA, it) }
      shouldAutoFallback?.let { putBoolean(SHOULD_AUTO_FALLBACK, it) }
      useServerStyles?.let { putBoolean(USE_SERVER_STYLES, it) }
      themeSetId?.let { putString(THEME_SET_ID, it) }
      routingCountry?.let { putString(ROUTING_COUNTRY, it) }
      locale?.let { putString(LOCALE, it) }
      consumeExceptions?.let { putBoolean(CONSUME_EXCEPTIONS, it) }
    }
  }

  /**
   * Contract to initiate and get the result from an Inquiry flow by using [registerForActivityResult].
   *
   * ```
   * val getInquiryResult = registerForActivityResult(Inquiry.Contract()) { result ->
   *   when(result) {
   *     is InquiryResponse.Complete -> {}
   *     is InquiryResponse.Cancel -> {}
   *     is InquiryResponse.Error -> {}
   *   }
   * }
   * ```
   */
  open class Contract
  @Deprecated("Use constructor Contract(Context) instead.")
  constructor() : ActivityResultContract<Inquiry, InquiryResponse>() {

    private var context: Context? = null

    @Suppress("DEPRECATION")
    constructor(context: Context) : this() {
      this.context = context
    }

    override fun createIntent(context: Context, input: Inquiry): Intent =
      input.toInquiryActivityIntent(context)

    override fun parseResult(resultCode: Int, intent: Intent?): InquiryResponse =
      @Suppress("DEPRECATION")
      onActivityResult(intent, context)
  }

  companion object {
    /**
     * Create an Inquiry flow builder based on a template ID.
     *
     * You can find your template ID on the Persona Dashboard under "Development.
     * https://withpersona.com/dashboard/integration
     *
     * @param templateId template ID from your Persona Dashboard
     * @return builder for the Inquiry flow
     */
    @JvmStatic
    fun fromTemplate(templateId: String): InquiryTemplateBuilder =
      InquiryTemplateBuilder(templateId = templateId)

    /**
     * Create an Inquiry flow builder based on a template ID version.
     *
     * You can find your template ID on the Persona Dashboard under "Development.
     * https://withpersona.com/dashboard/integration
     *
     * @param templateVersion template version from your Persona Dashboard
     * @return builder for the Inquiry flow
     */
    @JvmStatic
    fun fromTemplateVersion(templateVersion: String): InquiryTemplateBuilder =
      InquiryTemplateBuilder(templateVersion = templateVersion)

    /**
     * Create an Inquiry flow builder based on static template json.
     *
     * @param staticInquiryTemplate object that contains either a bundled resource id with linear steps
     * json or a template id for which remote linear steps json can be retrieved.
     * @return builder for the Inquiry flow
     */
    @ExperimentalInquiryFlow
    @JvmStatic
    fun fromStaticTemplate(staticInquiryTemplate: StaticInquiryTemplate): InquiryTemplateBuilder =
      InquiryTemplateBuilder(staticInquiryTemplate = staticInquiryTemplate)

    /**
     * Create an Inquiry flow builder based on an inquiry ID.
     *
     * You will need to generate the inquiry ID on the server. To try it out, you can create an
     * inquiry from the Persona Dashboard under "Inquiries". Click on the "Create Inquiry" button
     * and copy the inquiry ID from the URL.
     * https://withpersona.com/dashboard/inquiries
     *
     * @param inquiryId inquiry ID from your server
     * @return builder for the Inquiry flow
     */
    @JvmStatic
    fun fromInquiry(inquiryId: String): InquiryBuilder = InquiryBuilder(inquiryId)

    /**
     * Parse the response from the Persona Inquiry flow.
     *
     * @param intent from the [Activity.onActivityResult]
     * @return
     */
    @JvmStatic
    @JvmOverloads
    @MainThread
    @Deprecated("Use registerForActivityResult with the Inquiry.Contract instead.")
    fun onActivityResult(intent: Intent?, context: Context? = null): InquiryResponse =
      extractInquiryResponseFromBundle(intent?.extras, context)

    /**
     * Creates an [InquiryResponse] object from a result [Bundle] produced by the Inquiry SDK.
     */
    @JvmStatic
    fun extractInquiryResponseFromBundle(
      bundle: Bundle?,
      context: Context?,
    ): InquiryResponse {
      val status: Status = bundle.getStatus(context)
      val inquiryId = bundle?.getString(INQUIRY_ID)

      return when (status) {
        Status.INQUIRY_COMPLETE -> {
          val internalInquiryFields: InquiryFieldsMap =
            bundle?.let {
              BundleCompat.getParcelable(it, FIELDS_MAP, InquiryFieldsMap::class.java)
            } ?: InquiryFieldsMap(mapOf())
          val inquiryStatus = bundle?.getString(INQUIRY_STATUS)
          val collectedData: CollectedData? = bundle?.let {
            BundleCompat.getParcelable(it, COLLECTED_DATA, CollectedData::class.java)
          }

          InquiryResponse.Complete(
            inquiryId = inquiryId!!,
            status = inquiryStatus!!,
            fields = internalInquiryFields.fields.mapValues { (_, field) ->
              val inquiryField = when (field) {
                is InternalInquiryField.StringField -> StringField(field.value)
                is InternalInquiryField.IntegerField -> IntegerField(field.value)
                is InternalInquiryField.BooleanField -> BooleanField(field.value)
                is InternalInquiryField.FloatField -> FloatField(field.value)
                is InternalInquiryField.DateField -> {
                  field.value?.let {
                    val formatter = SimpleDateFormat("yyyy-MM-dd", Locale.US)
                    DateField(formatter.parse(it))
                  } ?: DateField(null)
                }
                is InternalInquiryField.DatetimeField -> {
                  field.value?.let {
                    // format datetime from ISO8601 string
                    val formatter = SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'", Locale.US)
                    DateField(formatter.parse(it))
                  } ?: DatetimeField(null)
                }
                is InternalInquiryField.ChoicesField -> ChoicesField(field.value)
                is InternalInquiryField.MultiChoicesField -> MultiChoicesField(field.value)
                is InternalInquiryField.Unknown -> UnknownField(field.type)
              }
              inquiryField
            },
            collectedData = collectedData,
          )
        }
        Status.INQUIRY_CANCELED -> {
          val sessionToken = bundle?.getString(SESSION_TOKEN)
          InquiryResponse.Cancel(
            inquiryId = inquiryId,
            sessionToken = sessionToken,
          )
        }
        Status.INQUIRY_ERROR -> {
          val debugErrorMessage = bundle?.getString(ERROR_DEBUG_MESSAGE)
          val lastErrorMessage = if (context != null) {
            ExceptionLogger(context).getLastErrorMessage()
          } else {
            null
          }
          val errorCode = try {
            bundle?.getParcelable(ERROR_CODE) as? ErrorCode ?: ErrorCode.UnexpectedError
          } catch (e: Exception) {
            ErrorCode.UnexpectedError
          }

          InquiryResponse.Error(
            debugMessage = debugErrorMessage ?: InquiryErrorMessages.unexpectedError,
            errorCode = errorCode,
            cause = lastErrorMessage,
          )
        }
      }
    }

    /**
     * Prefetches any models needed ahead of time for inquiries.
     */
    @Suppress("unused") // Public API
    fun prefetchModels() {
      Prefetching.prefetchModels()
    }

    /**
     * Cancels all running inquiries within this application.
     */
    @Suppress("unused") // Public API
    fun cancelRunningInquiries() {
      InquiryActivityBroadcastManager.cancelRunningInquiries()
    }

    private fun Bundle?.getStatus(context: Context?): Status {
      val status = this?.getString(PERSONA_ACTIVITY_RESULT)?.let {
        try {
          Status.valueOf(it)
        } catch (e: Exception) {
          null
        }
      }

      if (status != null) {
        return status
      }

      if (context == null) {
        return Status.INQUIRY_CANCELED
      }

      return if (ExceptionLogger(context).hasLastErrorMessage()) {
        Status.INQUIRY_ERROR
      } else {
        // This is most commonly caused by the app task getting killed and replaced. See PER-19177.
        Status.INQUIRY_CANCELED
      }
    }
  }
}