package com.withpersona.sdk2.inquiry

import androidx.annotation.StyleRes

/**
 * Get a this builder by calling [Inquiry.fromInquiry].
 */
class InquiryBuilder internal constructor(private val inquiryId: String) {
  private var theme: Int? = null
  private var themeSetId: String? = null
  private var sessionToken: String? = null
  private var useServerStyles: Boolean? = null
  private var routingCountry: String? = null
  private var locale: String? = null
  private var consumeExceptions: Boolean? = null

  /**
   * Theme to apply to the Inquiry flow
   *
   * @param theme a valid [StyleRes] to apply to an activity
   * @return
   */
  @Deprecated("Please migrate to theme(themeSource: ThemeSource) instead.")
  fun theme(@StyleRes theme: Int): InquiryBuilder = this.apply {
    this.theme = theme
    this.useServerStyles = false
  }

  /**
   * Theme to apply to the Inquiry flow
   *
   * @param themeSource a valid [ThemeSource] to apply for styling the inquiry
   * @return
   */
  fun theme(themeSource: ThemeSource): InquiryBuilder = this.apply {
    this.theme = themeSource.theme
    this.useServerStyles = themeSource is ServerThemeSource
  }

  /**
   * Theme set id to apply to the inquiry flow.
   *
   * @param themeSetId The themeSetId from which styles should be read from on the server.
   * @return
   */
  @ExperimentalThemeSetId
  @Suppress("unused") // Public API
  fun themeSetId(themeSetId: String): InquiryBuilder = this.apply {
    this.themeSetId = themeSetId
  }

  /**
   * Access token for resuming an [Inquiry]. The token must be generated on the server.
   *
   * @param sessionToken token required for resuming the [Inquiry]
   * @return
   */
  fun sessionToken(sessionToken: String?): InquiryBuilder =
    this.apply {
      if (!sessionToken.isNullOrEmpty()) {
        this.sessionToken = sessionToken
      }
    }

  /**
   * Set the country to route requests to directly. Only set this if you have been instructed to do so by the Persona team.
   * Requests may be re-routed and incur additional latency if this value is set incorrectly.
   * @param routingCountry The country containing server region to route requests to
   * @return
   */
  @Suppress("unused") // Public API
  fun routingCountry(routingCountry: String?): InquiryBuilder =
    this.apply { this.routingCountry = routingCountry }

  /**
   * Override the device locale with desired language.
   * @param locale A string in the form `ll` or `ll-RR` for language and region. e.g. `fr` or `fr-CA`.
   * @return
   */
  fun locale(locale: String?): InquiryBuilder =
    this.apply { this.locale = locale }

  /**
   * Set whether the SDK should consume fatal exceptions. When enabled, fatal exceptions caused by
   * the SDK will be consumed by the SDK and they will be returned as an error.
   *
   * Defaults to false.
   */
  @Suppress("unused") // Public API
  fun consumeExceptions(consumeExceptions: Boolean): InquiryBuilder =
    this.apply { this.consumeExceptions = consumeExceptions }

  /**
   * Create the [Inquiry] object with the fields set.
   *
   * @return
   */
  fun build(): Inquiry = Inquiry(
    templateId = null,
    templateVersion = null,
    inquiryId = inquiryId,
    sessionToken = sessionToken,
    referenceId = null,
    accountId = null,
    fields = null,
    theme = theme,
    themeSetId = themeSetId,
    environment = null,
    environmentId = null,
    enableErrorLogging = null,
    returnCollectedData = null,
    shouldAutoFallback = null,
    useServerStyles = useServerStyles,
    staticInquiryTemplate = null,
    routingCountry = routingCountry,
    locale = locale,
    consumeExceptions = consumeExceptions,
  )
}