package com.withpersona.sdk2.inquiry

import androidx.annotation.StyleRes

/**
 * Get a this builder by calling [Inquiry.fromTemplate] or [Inquiry.fromTemplateVersion].
 */
class InquiryTemplateBuilder internal constructor(
  private val templateId: String? = null,
  private val templateVersion: String? = null,
  private val staticInquiryTemplate: StaticInquiryTemplate? = null,
) {
  private var accountId: String? = null
  private var referenceId: String? = null
  private var fields: Fields? = null
  private var theme: Int? = null
  private var themeSetId: String? = null
  private var environment: Environment? = null
  private var environmentId: String? = null
  private var enableErrorLogging: Boolean? = null
  private var returnCollectedData: Boolean? = null
  private var shouldAutoFallback: Boolean? = null
  private var useServerStyles: Boolean? = null
  private var routingCountry: String? = null
  private var locale: String? = null
  private var consumeExceptions: Boolean? = null

  @Suppress("unused") // Public API
  fun fromTemplateId(templateId: String) = InquiryTemplateBuilder(templateId = templateId)

  /**
   * You can generate and provide a unique id which we will associate with the inquiry. The
   * identifier can be used to monitor user progress in newly created inquiries.
   *
   * Setting this will null out [accountId].
   *
   * @param referenceId custom identifier for the inquiry
   * @return
   */
  fun referenceId(referenceId: String?): InquiryTemplateBuilder = this.apply {
    this.accountId = null
    this.referenceId = referenceId
  }

  /**
   * Persona-maintained account used to monitor user progress across inquiries.
   *
   * Setting this will null out [referenceId].
   *
   * @param accountId custom identifier for the inquiry
   * @return
   */
  fun accountId(accountId: String?): InquiryTemplateBuilder = this.apply {
    this.referenceId = null
    this.accountId = accountId
  }

  /**
   * Pre-populate the Inquiry with custom fields
   *
   * @param fields custom data to store on the inquiry
   * @return
   */
  fun fields(fields: Fields?): InquiryTemplateBuilder = this.apply { this.fields = fields }

  /**
   * The Persona API environment on which to create inquiries. For sandbox and production, use
   * [Environment.SANDBOX] and [Environment.PRODUCTION] respectively.
   *
   * @param environment either [Environment.SANDBOX] or [Environment.PRODUCTION]
   * @return
   */
  fun environment(environment: Environment): InquiryTemplateBuilder =
    this.apply { this.environment = environment }

  /**
   * The specific Persona API environment on which to create inquiries. Use sandbox environments for development purposes only.
   *
   * @param environmentId is  in format env_1234567890
   * @return
   */
  fun environmentId(environmentId: String): InquiryTemplateBuilder =
    this.apply { this.environmentId = environmentId }

  /**
   * Theme to apply to the Inquiry flow
   *
   * @param theme a valid [StyleRes] to apply to an activity
   * @return
   */
  @Deprecated("Please migrate to theme(themeSource: ThemeSource) instead.")
  fun theme(@StyleRes theme: Int): InquiryTemplateBuilder = this.apply {
    this.theme = theme
    this.useServerStyles = false
  }

  /**
   * Theme to apply to the Inquiry flow
   *
   * @param themeSource a valid [ThemeSource] to apply for styling the inquiry
   * @return
   */
  fun theme(themeSource: ThemeSource): InquiryTemplateBuilder = this.apply {
    this.theme = themeSource.theme
    this.useServerStyles = themeSource is ServerThemeSource
  }

  /**
   * Theme set id to apply to the inquiry flow.
   *
   * @param themeSetId The themeSetId from which styles should be read from on the server.
   * @return
   */
  @ExperimentalThemeSetId
  @Suppress("unused") // Public API
  fun themeSetId(themeSetId: String): InquiryTemplateBuilder = this.apply {
    this.themeSetId = themeSetId
  }

  /**
   * Whether to enable error logging for fatal exceptions. When enabled, additional information will
   * be available in the [InquiryResponse.Error] object when an error occurs due to an uncaught
   * exception. See [InquiryResponse.Error.cause].
   *
   * Defaults to true.
   */
  @Suppress("unused") // Public API
  fun enableErrorLogging(enableErrorLogging: Boolean): InquiryTemplateBuilder =
    this.apply { this.enableErrorLogging = enableErrorLogging }

  /**
   * Whether to return the data collected from the inquiry process.
   *
   * Defaults to false.
   */
  fun returnCollectedData(returnCollectedData: Boolean): InquiryTemplateBuilder =
    this.apply { this.returnCollectedData = returnCollectedData }

  /**
   * Whether to enable auto fallback mode. When enabled, the inquiry will automatically check fallback
   * system for given template. Work with your account team to enable this feature.
   *
   * Defaults to false.
   */
  @ExperimentalFallbackMode
  fun shouldAutoFallback(shouldAutoFallback: Boolean): InquiryTemplateBuilder =
    this.apply { this.shouldAutoFallback = shouldAutoFallback }

  /**
   * Set the country to route requests to directly. Only set this if you have been instructed to do so by the Persona team.
   * Requests may be re-routed and incur additional latency if this value is set incorrectly.
   * @param routingCountry The country containing server region to route requests to. e.g. RoutingCountry.DE.name
   * @return
   */
  @Suppress("unused") // Public API
  fun routingCountry(routingCountry: String?): InquiryTemplateBuilder =
    this.apply { this.routingCountry = routingCountry }

  /**
   * Override the device locale with desired language.
   * @param locale A string in the form `ll` or `ll-RR` for language and region. e.g. `fr` or `fr-CA`.
   * @return
   */
  fun locale(locale: String?): InquiryTemplateBuilder =
    this.apply { this.locale = locale }

  /**
   * Set whether the SDK should consume fatal exceptions. When enabled, fatal exceptions caused by
   * the SDK will be consumed by the SDK and they will be returned as an error.
   *
   * Defaults to false.
   */
  @Suppress("unused") // Public API
  fun consumeExceptions(consumeExceptions: Boolean): InquiryTemplateBuilder =
    this.apply { this.consumeExceptions = consumeExceptions }

  /**
   * Create the [Inquiry] objects with the fields set.
   *
   * @return
   */
  fun build(): Inquiry = Inquiry(
    templateId = templateId,
    templateVersion = templateVersion,
    inquiryId = null,
    sessionToken = null,
    referenceId = referenceId,
    accountId = accountId,
    fields = fields,
    theme = theme,
    themeSetId = themeSetId,
    environment = environment,
    environmentId = environmentId,
    enableErrorLogging = enableErrorLogging,
    returnCollectedData = returnCollectedData,
    shouldAutoFallback = shouldAutoFallback,
    useServerStyles = useServerStyles,
    staticInquiryTemplate = staticInquiryTemplate,
    routingCountry = routingCountry,
    locale = locale,
    consumeExceptions = consumeExceptions,
  )
}