package com.withpersona.sdk2.inquiry.error_reporting

import android.content.Context
import java.io.ByteArrayOutputStream
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.io.OutputStreamWriter
import java.nio.charset.Charset

/**
 * Handles logging an exception. It only supports logging one exception at a time. If more than one
 * exception is logged, only the newest exception will be kept.
 *
 * The exception handler is loosely based off of firebase crashlytics.
 * https://github.com/firebase/firebase-android-sdk/tree/master/firebase-crashlytics
 */
class ExceptionLogger(context: Context) {

  companion object {
    private const val SDK_FILES_FOLDER_NAME = ".com.withpersona.sdk2.inquiry"
    private const val CRASH_LOG_FOLDER_NAME = "errors"

    private val UTF_8 = Charset.forName("UTF-8")
  }

  private val errDir: File

  init {
    // based on
    // https://github.com/firebase/firebase-android-sdk/blob/master/firebase-crashlytics/src/main/java/com/google/firebase/crashlytics/internal/persistence/FileStore.java
    val baseDir = context.filesDir
    val sdkDir = File(baseDir, SDK_FILES_FOLDER_NAME)
    errDir = File(sdkDir, CRASH_LOG_FOLDER_NAME)
  }

  /**
   * Logs the given [Throwable] to disk.
   */
  fun logException(error: Throwable) {
    prepareDir(errDir)

    val errorFile = getLastErrorFile()
    if (errorFile.isDirectory) {
      errorFile.delete()
    }

    writeTextFile(
      file = errorFile,
      text = "${error.javaClass.canonicalName} in ${error.findRelevantCallInCallStack() ?: error.firstFromCallStack()}",
    )
  }

  /**
   * Deletes the last error log.
   */
  fun clearLastError() {
    getLastErrorFile().delete()
  }

  /**
   * @return the last error log.
   */
  fun getLastErrorMessage(): String? =
    readTextFile(getLastErrorFile())

  fun hasLastErrorMessage(): Boolean =
    getLastErrorFile().exists()

  private fun getLastErrorFile() =
    File(errDir, "last_error.txt")

  /**
   * @return the first call found of a method that is from our code base from a [Throwable]. Returns
   * null if no such call is found.
   */
  private fun Throwable.findRelevantCallInCallStack(): String? {
    var currentError: Throwable? = this
    while (currentError != null) {
      currentError.stackTrace.forEach {
        if (it.className.startsWith("com.withpersona.sdk2.inquiry")) {
          return "${it.className} in ${it.methodName} at line ${it.lineNumber}"
        }
      }
      currentError = currentError.cause
    }
    return null
  }

  /**
   * @return the first call of a [Throwable] as a human readable string.
   */
  private fun Throwable.firstFromCallStack(): String? =
    this.stackTrace.firstOrNull()?.let {
      "${it.className} in ${it.methodName} at line ${it.lineNumber}"
    }

  /**
   * Create the directory referred by [File]. If the given [File] exist but is not a directory,
   * the [File] will be deleted before the directory is created.
   */
  private fun prepareDir(file: File) {
    if (file.exists()) {
      if (file.isDirectory) {
        return
      } else {
        file.delete()
      }
    }
    file.mkdirs()
  }

  // Inspired by
  // https://github.com/firebase/firebase-android-sdk/blob/master/firebase-crashlytics/src/main/java/com/google/firebase/crashlytics/internal/persistence/CrashlyticsReportPersistence.java
  private fun readTextFile(file: File): String? {
    if (!file.exists() || !file.isFile) {
      return null
    }

    val readBuffer = ByteArray(8192)
    val bos = ByteArrayOutputStream()
    FileInputStream(file).use { fileInput ->
      var read: Int
      while (fileInput.read(readBuffer).also { read = it } > 0) {
        bos.write(readBuffer, 0, read)
      }
      return String(bos.toByteArray(), UTF_8)
    }
  }

  // Inspired by
  // https://github.com/firebase/firebase-android-sdk/blob/master/firebase-crashlytics/src/main/java/com/google/firebase/crashlytics/internal/persistence/CrashlyticsReportPersistence.java
  private fun writeTextFile(file: File, text: String) {
    OutputStreamWriter(FileOutputStream(file), UTF_8).use { writer ->
      writer.write(text)
    }
  }
}