package com.withpersona.sdk2.inquiry.internal

import android.os.Parcelable
import com.squareup.moshi.FromJson
import com.squareup.moshi.JsonAdapter
import com.squareup.moshi.JsonClass
import com.squareup.moshi.JsonReader
import com.squareup.moshi.JsonWriter
import com.squareup.moshi.ToJson
import com.squareup.moshi.adapters.PolymorphicJsonAdapterFactory
import kotlinx.parcelize.Parcelize

class InquiryFieldMap(
  val fields: Map<String, InquiryField>,
) {
  companion object : JsonAdapter<InquiryFieldMap>() {
    @FromJson
    override fun fromJson(reader: JsonReader): InquiryFieldMap? {
      TODO("Not yet implemented")
    }

    @ToJson
    override fun toJson(writer: JsonWriter, value: InquiryFieldMap?) {
      if (value == null) {
        writer.nullValue()
        return
      }

      writer.beginObject()
      value.fields.forEach { (name, value) ->
        writer.name(name)
        when (value) {
          is InquiryField.StringField -> writer.value(value.value)
          is InquiryField.IntegerField -> writer.value(value.value)
          is InquiryField.BooleanField -> writer.value(value.value)
          is InquiryField.DatetimeField -> writer.value(value.value)
          is InquiryField.DateField -> writer.value(value.value)
          is InquiryField.FloatField -> writer.value(value.value)
          is InquiryField.ChoicesField -> writer.value(value.value)
          is InquiryField.MultiChoicesField -> {
            writer.beginArray()
            value.value?.forEach { writer.value(it) }
            writer.endArray()
          }
          is InquiryField.Unknown ->
            // this should never happen
            error("Attempted to write field with type `Unknown`.")
        }
      }
      writer.endObject()
    }
  }
}

sealed class InquiryField(
  open val type: String,
) : Parcelable {
  companion object {
    fun createAdapter(): JsonAdapter.Factory =
      PolymorphicJsonAdapterFactory.of(InquiryField::class.java, "type")
        .withFallbackJsonAdapter(UnknownAdapter as JsonAdapter<Any>)
        .withSubtype(StringField::class.java, StringField.TYPE)
        .withSubtype(IntegerField::class.java, IntegerField.TYPE)
        .withSubtype(BooleanField::class.java, BooleanField.TYPE)
        .withSubtype(FloatField::class.java, FloatField.TYPE)
        // Server returns float fields back to us as "number" fields.
        .withSubtype(FloatField::class.java, FloatField.TYPE2)
        .withSubtype(DateField::class.java, DateField.TYPE)
        .withSubtype(DatetimeField::class.java, DatetimeField.TYPE)
        .withSubtype(ChoicesField::class.java, ChoicesField.TYPE)
        .withSubtype(MultiChoicesField::class.java, MultiChoicesField.TYPE)
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class StringField(
    val value: String?,
    override val type: String = StringField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "string"
    }
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class IntegerField(
    val value: Int?,
    override val type: String = IntegerField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "integer"
    }
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class BooleanField(
    val value: Boolean?,
    override val type: String = BooleanField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "boolean"
    }
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class FloatField(
    val value: Float?,
    override val type: String = FloatField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "float"
      const val TYPE2: String = "number"
    }
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class DateField(
    val value: String?,
    override val type: String = DateField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "date"
    }
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class DatetimeField(
    val value: String?,
    override val type: String = DatetimeField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "datetime"
    }
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class ChoicesField(
    val value: String?,
    override val type: String = ChoicesField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "choices"
    }
  }

  @JsonClass(generateAdapter = true)
  @Parcelize
  data class MultiChoicesField(
    val value: Array<String>?,
    override val type: String = MultiChoicesField.TYPE,
  ) : InquiryField(type) {
    companion object {
      const val TYPE: String = "multi_choices"
    }

    override fun equals(other: Any?): Boolean {
      if (this === other) return true
      if (javaClass != other?.javaClass) return false

      other as MultiChoicesField

      if (value != null) {
        if (other.value == null) return false
        if (!value.contentEquals(other.value)) return false
      } else if (other.value != null) return false

      return true
    }

    override fun hashCode(): Int {
      return value?.contentHashCode() ?: 0
    }
  }

  @Parcelize
  class Unknown(override val type: String) : InquiryField(type)

  private object UnknownAdapter : JsonAdapter<InquiryField>() {
    override fun toJson(writer: JsonWriter, value: InquiryField?) {
      writer.beginObject()
      writer.name("type")
      writer.value(value?.type)
      writer.endObject()
    }

    override fun fromJson(reader: JsonReader): InquiryField {
      var type = ""
      reader.beginObject()
      while (reader.hasNext()) {
        val name = reader.nextName()
        if (name == "type") {
          type = reader.nextString()
        } else {
          reader.skipValue()
        }
      }
      reader.endObject()
      return Unknown(type)
    }
  }
}